from ._version import __version__
from .base import MedicalImage
from fileformats.application import Dicom  # imported to alias it here as well
from .misc import (  # noqa: F401
    Analyze,
    Mgh,
    MghGz,
)
from .nifti import (
    Nifti,
    Nifti1,
    Nifti2,
    NiftiGz,
    NiftiX,
    NiftiGzX,
)
from .diffusion import (
    DwiEncoding,
    Bvec,
    Bval,
    NiftiBvec,
    NiftiGzBvec,
    NiftiXBvec,
    NiftiGzXBvec,
)
from .dicom import (  # noqa: F401
    DicomCollection,
    DicomDir,
    DicomSeries,
    # SiemensDicomDir,
)
from .raw import (  # noqa: F401
    Kspace,
    Rda,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
)



import typing as ty
from fileformats.core import hook
from fileformats.core.mixin import WithAdjacentFiles
from fileformats.generic import File
from .nifti import NiftiGzX, NiftiGz, Nifti1, NiftiX


class DwiEncoding(File):

    iana_mime: ty.Optional[str] = None

    @hook.extra
    def read_array(self):
        "Both the gradient direction and weighting combined into a single Nx4 array"
        raise NotImplementedError

    @property
    def array(self):
        return self.read_array()

    @property
    def directions(self):
        "gradient direction and weighting combined into a single Nx4 array"
        return self.array[:, :3]

    @property
    def b_values(self):
        "the b-value weighting"
        return self.array[:, 3]


class Bval(File):

    ext = ".bval"

    @hook.extra
    def read_array(self):
        raise NotImplementedError


class Bvec(WithAdjacentFiles, DwiEncoding):
    """FSL-style diffusion encoding, in two separate files"""

    ext = ".bvec"

    @hook.required
    @property
    def b_values_file(self) -> Bval:
        return Bval(self.select_by_ext(Bval))


# NIfTI file format gzipped with BIDS side car
class WithBvec(WithAdjacentFiles):
    @hook.required
    @property
    def encoding(self) -> Bvec:
        return Bvec(self.select_by_ext(Bvec))


class NiftiBvec(WithBvec, Nifti1):
    iana_mime = "application/x-nifti2+bvec"


class NiftiGzBvec(WithBvec, NiftiGz):
    iana_mime = "application/x-nifti2+gzip.bvec"


class NiftiXBvec(WithBvec, NiftiX):
    iana_mime = "application/x-nifti2+json.bvec"


class NiftiGzXBvec(WithBvec, NiftiGzX):
    iana_mime = "application/x-nifti2+gzip.json.bvec"



from pathlib import Path
import typing as ty
import nibabel
from fileformats.core import FileSet
from fileformats.medimage import MedicalImage, Nifti, NiftiGz, Nifti1, NiftiGzX, NiftiX
import medimages4tests.dummy.nifti


@FileSet.read_metadata.register
def nifti_read_metadata(nifti: Nifti):
    return dict(nibabel.load(nifti.fspath).header)


@MedicalImage.read_array.register
def nifti_data_array(nifti: Nifti):
    return nibabel.load(nifti.fspath).get_data()


@MedicalImage.vox_sizes.register
def nifti_vox_sizes(nifti: Nifti):
    # FIXME: This won't work for 4-D files
    return nifti.metadata["pixdim"][1:4]


@MedicalImage.dims.register
def nifti_dims(nifti: Nifti):
    # FIXME: This won't work for 4-D files
    return nifti.metadata["dim"][1:4]


@FileSet.generate_sample_data.register
def nifti_generate_sample_data(nifti: Nifti1, dest_dir: Path, seed: int, stem: ty.Optional[str]):
    return medimages4tests.dummy.nifti.get_image(out_file=dest_dir / "nifti.nii")


@FileSet.generate_sample_data.register
def nifti_gz_generate_sample_data(nifti: NiftiGz, dest_dir: Path, seed: int, stem: ty.Optional[str]):
    return medimages4tests.dummy.nifti.get_image(
        out_file=dest_dir / "nifti.nii.gz", compressed=True
    )


@FileSet.generate_sample_data.register
def nifti_gz_x_generate_sample_data(nifti: NiftiGzX, dest_dir: Path, seed: int, stem: ty.Optional[str]):
    return medimages4tests.mri.neuro.t1w.get_image()


@FileSet.generate_sample_data.register
def nifti_x_generate_sample_data(nifti: NiftiX, dest_dir: Path, seed: int, stem: ty.Optional[str]):
    nifti_gz_x = NiftiGzX(medimages4tests.mri.neuro.t1w.get_image())
    return NiftiX.convert(nifti_gz_x)



from pathlib import Path
import typing as ty
from tempfile import mkdtemp
from random import Random
import pydicom
import numpy as np
from fileformats.core import FileSet
from fileformats.core.utils import gen_filename
from fileformats.medimage import MedicalImage, DicomCollection, DicomDir, DicomSeries
import medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c


@MedicalImage.read_array.register
def dicom_read_array(collection: DicomCollection):
    image_stack = []
    for dcm_file in collection.contents:
        image_stack.append(pydicom.dcmread(dcm_file).pixel_array)
    return np.asarray(image_stack)


@MedicalImage.vox_sizes.register
def dicom_vox_sizes(collection: DicomCollection):
    return tuple(
        collection.metadata["PixelSpacing"] + [collection.metadata["SliceThickness"]]
    )


@MedicalImage.dims.register
def dicom_dims(collection: DicomCollection):
    return tuple(
        (
            collection.metadata["Rows"],
            collection.metadata["DataColumns"],
            len(list(collection.contents)),
        ),
    )


@DicomCollection.series_number.register
def dicom_series_number(collection: DicomCollection):
    return str(collection.metadata["SeriesNumber"])


@FileSet.generate_sample_data.register
def dicom_dir_generate_sample_data(dcmdir: DicomDir, dest_dir: Path, seed: ty.Union[int, Random], stem: ty.Optional[str]):
    dcm_dir = medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c.get_image()
    # Set series number to random value to make it different
    if isinstance(seed, Random):
        rng = seed
    else:
        rng = Random(seed)
    series_number = rng.randint(1, SERIES_NUMBER_RANGE)
    dest = Path(dest_dir) / gen_filename(seed_or_rng=seed, stem=stem)
    dest.mkdir()
    for dcm_file in dcm_dir.iterdir():
        dcm = pydicom.dcmread(dcm_file)
        dcm.SeriesNumber = series_number
        pydicom.dcmwrite(dest / dcm_file.name, dcm)
    return [dest]


@FileSet.generate_sample_data.register
def dicom_set_generate_sample_data(dcm_series: DicomSeries, dest_dir: Path, seed: int, stem: ty.Optional[str]):
    rng = Random(seed)
    dicom_dir = dicom_dir_generate_sample_data(dcm_series, dest_dir=mkdtemp(), seed=rng, stem=None)[0]
    stem = gen_filename(rng, stem=stem)
    fspaths = []
    for i, dicom_file in enumerate(dicom_dir.iterdir(), start=1):
        fspaths.append(dicom_file.rename(dest_dir / f"{stem}-{i}.dcm"))
    return fspaths


SERIES_NUMBER_TAG = ("0020", "0011")
SERIES_NUMBER_RANGE = int(1e8)



import typing as ty
import logging
from fileformats.generic import FileSet
from fileformats.core import hook

logger = logging.getLogger("fileformats")


# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class MedicalImage(FileSet):

    iana_mime: ty.Optional[str] = None
    INCLUDE_HDR_KEYS = None
    IGNORE_HDR_KEYS = None
    binary = True

    @hook.extra
    def read_array(self):
        """
        Returns the binary data of the image in a numpy array
        """
        raise NotImplementedError

    @hook.extra
    def vox_sizes(self) -> ty.Tuple[float]:
        """The length of the voxels along each dimension"""
        raise NotImplementedError

    @hook.extra
    def dims(self) -> ty.Tuple[int]:
        """The dimensions of the image"""
        raise NotImplementedError



from fileformats.medimage import DicomSeries, DicomDir


def test_dicom_series_metadata(tmp_path):
    series = DicomSeries.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series["SOPInstanceUID"], list)


def test_dicom_dir_metadata(tmp_path):
    series = DicomDir.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series["SOPInstanceUID"], list)



import typing as ty
from copy import copy
from operator import itemgetter
from collections import defaultdict
from pathlib import Path
from functools import cached_property
from fileformats.core import hook, FileSet
from fileformats.generic import DirectoryContaining, SetOf
from fileformats.application import Dicom
from .base import MedicalImage

# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class DicomCollection(MedicalImage):
    """Base class for collections of DICOM files, which can either be stored within a
    directory (DicomDir) or presented as a flat list (DicomSeries)
    """

    content_types = (Dicom,)
    iana_mime = None

    def __len__(self):
        return len(self.contents)

    @hook.extra
    def series_number(self):
        raise NotImplementedError

    @cached_property
    def contents(self) -> ty.List[Dicom]:
        return sorted(super().contents, key=itemgetter("SOPInstanceUID"))


class DicomDir(DicomCollection, DirectoryContaining[Dicom]):
    pass


class DicomSeries(DicomCollection, SetOf[Dicom]):
    @classmethod
    def from_paths(
        cls, fspaths: ty.Iterable[Path], common_ok: bool = False
    ) -> ty.Tuple[ty.Set[FileSet], ty.Set[Path]]:
        dicoms, remaining = Dicom.from_paths(fspaths, common_ok=common_ok)
        series_dict = defaultdict(list)
        for dicom in dicoms:
            series_dict[(str(dicom["StudyInstanceUID"]), str(dicom["SeriesNumber"]))].append(dicom)
        return set([cls(s) for s in series_dict.values()]), remaining


@FileSet.read_metadata.register
def dicom_collection_read_metadata(collection: DicomCollection) -> ty.Dict[str, ty.Any]:
    # Collated DICOM headers across series
    collated = copy(collection.contents[0].metadata)
    for i, dicom in enumerate(collection.contents[1:], start=1):
        for key, val in dicom.metadata.items():
            if val != collated[key]:
                # Check whether the value is the same as the values in the previous
                # images in the series
                if (
                    not isinstance(collated[key], list)
                    or isinstance(val, list)
                    and not isinstance(collated[key][0], list)
                ):
                    collated[key] = [collated[key]] * i + [val]
                collated[key].append(val)
    return collated



from pathlib import Path
import attrs
import json
import typing as ty
import tempfile
import jq
from fileformats.core import hook
import pydra
from fileformats.medimage.base import MedicalImage
from fileformats.medimage.dicom import DicomDir, DicomCollection, DicomSeries
from fileformats.medimage import (
    Analyze,
    Nifti,
    NiftiGz,
    NiftiX,
    NiftiGzX,
    NiftiXBvec,
    NiftiBvec,
    NiftiGzBvec,
    NiftiGzXBvec,
)

try:
    from pydra.tasks.mrtrix3.utils import MRConvert
except ImportError:
    from pydra.tasks.mrtrix3.latest import mrconvert as MRConvert
from pydra.tasks.dcm2niix import Dcm2Niix


@hook.converter(source_format=MedicalImage, target_format=Analyze, out_ext=Analyze.ext)
def mrconvert(name, out_ext: str):
    """Initiate an MRConvert task with the output file extension set

    Parameters
    ----------
    name : str
        name of the converter task
    out_ext : str
        extension of the output file, used by MRConvert to determine the desired format

    Returns
    -------
    pydra.ShellCommandTask
        the converter task
    """
    return MRConvert(name=name, out_file="out" + out_ext)


@pydra.mark.task
def ensure_dicom_dir(dicom: DicomCollection) -> DicomDir:
    if isinstance(dicom, DicomSeries):
        dicom_dir_fspath = tempfile.mkdtemp()
        dicom.copy(dicom_dir_fspath, mode=DicomDir.CopyMode.link)
        dicom = DicomDir(dicom_dir_fspath)
    elif not isinstance(dicom, DicomDir):
        raise RuntimeError(
            "Unrecognised input to ensure_dicom_dir, should be DicomSeries or DicomDir "
            f"not {dicom}"
        )
    return dicom


@hook.converter(source_format=DicomCollection, target_format=Nifti)
@hook.converter(source_format=DicomCollection, target_format=NiftiGz, compress="y")
@hook.converter(source_format=DicomCollection, target_format=NiftiX)
@hook.converter(source_format=DicomCollection, target_format=NiftiGzX, compress="y")
@hook.converter(source_format=DicomCollection, target_format=NiftiXBvec)
@hook.converter(source_format=DicomCollection, target_format=NiftiBvec)
@hook.converter(source_format=DicomCollection, target_format=NiftiGzBvec)
@hook.converter(
    source_format=DicomCollection, target_format=NiftiGzXBvec, compress="y"
)
def extended_dcm2niix(
    name,
    compress: str = "n",
    file_postfix: ty.Optional[str] = None,
    side_car_jq: ty.Optional[str] = None,
    extract_volume: ty.Optional[int] = None,
    to_4d: bool = False,
):
    """The Dcm2niix command wrapped in a workflow in order to map the inputs and outputs
    onto "in_file" and "out_file", respectively, and implement optional post-conversion
    manipulations to allow manual override of conversion issues.

    Parameters
    ----------
    name : str
        name of the workflow
    compress : str, optional
        whether to apply compression to the conversion, by default "n"
    file_postfix : str, optional
        select one of the multiple output files by its generated postfix.
        See https://github.com/rordenlab/dcm2niix/blob/master/FILENAMING.md for
        complete list of different postfixes that will be generated, by default None
    side_car_jq : str, optional
        JQ (https://stedolan.github.io/jq/) expression that can be provided to edit
        the resulting JSON side-car. Can be used to fix any conversion issues or add
        required fields manually, by default None
    extract_volume : int, optional
        extract a 3D volume from a 4D dataset by passing the index (0-based) of the
        volume to extract, by default None
    to_4d : bool, optional
        whether to wrap resulting 3D NIfTI volume in a 4D dataset, by default False

    Returns
    -------
    pydra.Workflow
        the converter workflow

    Raises
    ------
    ValueError
        when mutually exclusive "extract_volume" and "to_4d" options are provided
    """

    if extract_volume is not None and to_4d:
        raise ValueError(
            f"'extract_volume' ({extract_volume}) and 'to_4d' are mutually exclusive"
        )
    # Create workflow to map input field to "in_file" and optionally perform post-conversion
    # steps to manipulate the converted NIfTI files
    wf = pydra.Workflow(
        name=name,
        input_spec=["in_file"],
    )
    wf.add(
        ensure_dicom_dir(
            dicom=wf.lzin.in_file,
            name="ensure_dicom_dir",
        )
    )

    if file_postfix is None:
        file_postfix = attrs.NOTHING  # type: ignore
    wf.add(
        Dcm2Niix(
            in_dir=wf.ensure_dicom_dir.lzout.out,
            out_dir=Path("."),
            name="dcm2niix",
            compress=compress,
            file_postfix=file_postfix,
        )
    )
    out_file = wf.dcm2niix.lzout.out_file
    out_json = wf.dcm2niix.lzout.out_json
    # Add MRConvert step to either select a single volume of a 4D dataset or the inverse,
    # wrap a single volume in a 4D dataset
    if extract_volume is not None or to_4d:
        if extract_volume:
            coord = [3, extract_volume]
            axes = [0, 1, 2]
        else:  # to_4d
            coord = attrs.NOTHING  # type: ignore
            axes = [0, 1, 2, -1]
        wf.add(
            MRConvert(
                in_file=out_file,
                coord=coord,
                axes=axes,
                name="mrconvert",
            )
        )
        out_file = wf.mrconvert.lzout.out_file
    # Add JQ edit of side car to allow manual fixing of any conversion issues
    if side_car_jq is not None:
        wf.add(
            edit_dcm2niix_side_car(
                in_file=out_json, jq_expr=side_car_jq, name="json_edit"
            )
        )
        out_json = wf.json_edit.lzout.out

    wf.add(
        collect_dcm2niix_outputs(
            name="collect_outputs",
            out_file=out_file,
            out_json=out_json,
            out_bvec=wf.dcm2niix.lzout.out_bvec,
            out_bval=wf.dcm2niix.lzout.out_bval,
        )
    )
    # Set workflow outputs
    wf.set_output(("out_file", wf.collect_outputs.lzout.out))
    return wf


@pydra.mark.task
def edit_dcm2niix_side_car(in_file: Path, jq_expr: str, out_file=None) -> Path:
    """ "Applies ad-hoc edit of JSON side car with JQ query language"""
    if out_file is None:
        out_file = in_file
    with open(in_file) as f:
        dct = json.load(f)
    dct = jq.compile(jq_expr).input(dct).first()
    with open(out_file, "w") as f:
        json.dump(dct, f)
    return in_file


@pydra.mark.task
def collect_dcm2niix_outputs(
    out_file: Path, out_json: Path, out_bvec: Path, out_bval: Path
) -> ty.List[Path]:
    lst = [out_file]
    for file in (out_json, out_bvec, out_bval):
        if file is not attrs.NOTHING:
            lst.append(file)
    return lst



from .base import (
    PetRawData,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
)


class Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData(PetRawData):
    iana_mime = None
    ext = ".ptd"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode(
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData, PetListMode
):
    pass


class Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram(
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData, PetSinogram
):
    "histogrammed projection data in a reconstruction-friendly format"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate(
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData, PetCountRate
):
    "number of prompt/random/single events per unit time"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation(
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData, PetNormalisation
):
    "normalisation scan or the current cross calibration factor"



import numpy as np
from fileformats.medimage.diffusion import DwiEncoding, Bval, Bvec


@Bval.read_array.register
def bval_read_array(bval: Bval):
    return np.asarray([float(ln) for ln in bval.read_contents().split()])


@DwiEncoding.read_array.register
def bvec_read_array(bvec: Bvec):
    bvals = bvec.b_values_file.read_array()
    directions = np.asarray(
        [[float(x) for x in ln.split()] for ln in bvec.read_contents().splitlines()]
    ).T
    return np.concatenate((directions, bvals.reshape((-1, 1))), axis=1)  # type: ignore
